// File: Task.java
// Construction of a task, for the algorithms described in:
// Hamed Fahimi and Claude-Guy Quimper,
// "Linear-Time Filtering Algorithms for the Disjunctive Constraint", AAAI-2014.
// By: Hamed Fahimi



import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

class Task implements Comparable<Task> {
    private int est;
    private int lct;
    private int lst;
    private int ect;
    private int p;
    public int est_to_timepoint_index;
    public int ect_to_timepoint_index;
    public int lct_to_timepoint_index;
    public int lst_to_timepoint_index;

    public static class ComparatorByEst implements Comparator<Integer> {

        private  Task[] tasks;

            public ComparatorByEst( Task[] list_of_tasks) {
 
            this.tasks = list_of_tasks;
        }
        @Override
        public int compare(Integer a, Integer b) {
         return tasks[a].earliestStartingTime() - tasks[b].earliestStartingTime();
        }
    }

    public static class ComparatorByLct implements Comparator<Integer> {
        private List<Task> tasks;

        public ComparatorByLct(List<Task> list_of_tasks) {
            this.tasks = list_of_tasks;
        }
        @Override
        public int compare(Integer a, Integer b) {
            return tasks.get(a.intValue()).latestCompletionTime() - tasks.get(b.intValue()).latestCompletionTime();
        }
    }

    public static class ComparatorByEct implements Comparator<Integer> {
        private Task[] tasks;
        public ComparatorByEct(Task[] list_of_tasks) {
            this.tasks = list_of_tasks;
        }
        @Override
        public int compare(Integer a, Integer b) {
            return tasks[a].earliestCompletionTime() - tasks[b].earliestCompletionTime();
        }
    }

    public static class ComparatorByLst implements Comparator<Integer> {
        private  Task[] tasks;

            public ComparatorByLst( Task[] list_of_tasks) {
  
            this.tasks = list_of_tasks;
        }
        @Override
        public int compare(Integer a, Integer b) {
            return tasks[a].latestStartingTime() - tasks[b].latestStartingTime();

        }
    }
    
    
    public static class ComparatorByP implements Comparator<Integer> {
        private List<Task> tasks;
        public ComparatorByP(List<Task> list_of_tasks) {
            this.tasks = list_of_tasks;
        }
        @Override
        public int compare(Integer a, Integer b) {
            return tasks.get(a.intValue()).processingTime() - tasks.get(b.intValue()).processingTime();
        }
    }
    
    
    
    

    public static class TaskByest implements Comparator<Task> {
        @Override
        public int compare(Task o1, Task o2) {
            return o1.earliestStartingTime() > o2.earliestStartingTime() ? 1 : (o1.earliestStartingTime() < o2.earliestStartingTime() ? -1 : 0);
        }
    }


    public static class TaskBylct implements Comparator<Task> {
        @Override
        public int compare(Task o1, Task o2) {
            return o1.latestCompletionTime() > o2.latestCompletionTime() ? 1 : (o1.latestCompletionTime() < o2.latestCompletionTime() ? -1 : 0);
        }
    }

    public static class TaskBylst implements Comparator<Task> {
        @Override
        public int compare(Task o1, Task o2) {
            return o1.latestStartingTime() > o2.latestStartingTime() ? 1 : (o1.latestStartingTime() < o2.latestStartingTime() ? -1 : 0);
        }
    }

    public static class TaskByect implements Comparator<Task> {
        @Override
        public int compare(Task o1, Task o2) {
            return o1.earliestCompletionTime() > o2.earliestCompletionTime() ? 1 : (o1.earliestCompletionTime() < o2.earliestCompletionTime() ? -1 : 0);
        }
    }
    
    public static class TaskByprocessingtime implements Comparator<Task> {
        @Override
        public int compare(Task o1, Task o2) {
            return o1.processingTime() > o2.processingTime() ? 1 : (o1.processingTime() < o2.processingTime() ? -1 : 0);
        }
    }

    public Task(int est, int lct, int p) {
        this.est = est;
        this.lct = lct;
        this.p = p;
    }

    public int earliestStartingTime() {
        return est;
    }

    public int latestCompletionTime() {
        return lct;
    }

    public int processingTime() {
        return p;
    }

    public int earliestCompletionTime() {
        return est + p;
    }

    public int latestStartingTime() {
        return lct - p;
    }

    
    
    public void setEarliestStartingTime(int est) {
        this.est = est;
    }  
    
    public void setLatestStartingTime(int lst) {
        this.lst = lst;
    }  
   
    public void setEarliestCompletionTime(int ect) {
        this.ect = ect;
    } 
    
    public void setLatestCompletionTime(int lct) {
        this.lct = lct;
    } 
    
    
    public void setProcessingTime(int p) {
        this.p = p;
    }
    
    
    public boolean pruneEarliestStartingTime(int newest) {
        if (newest > est) {
            est = newest;
            return true;
        }
	return false;
    }

    public boolean pruneLatestCompletionTime(Integer newlct) {
        if ((newlct != null) && (newlct < lct)) {
            lct = newlct;
            return true;
        } else {
            return false;
        }
    }    

    public void fixedPart() {
        if (latestStartingTime() < earliestCompletionTime()) {
            System.out.printf("The fixed part is:(%d", latestStartingTime(), " , %d", earliestCompletionTime(), ")");
            System.out.printf(" , %d", earliestCompletionTime());
            System.out.printf(")\n");
        } else {
            System.out.printf("No fixed part!");
        }
    }

    public boolean isConsistent() {
        return (earliestCompletionTime() <= latestCompletionTime());
    }
    
    public boolean hasFixedPart() {
        return est + 2 * p > lct;
    }
    
    // Returns true if the task has a fixed part that overlaps with the fixed part of another task.
    public boolean haveOverlappingFixedParts(Task another_task) {
        
        return this.hasFixedPart() && another_task.hasFixedPart() && overlap(this.latestStartingTime(),
                this.earliestCompletionTime(),
                another_task.latestStartingTime(),
                another_task.earliestCompletionTime());
        
        
        
    }
    
    private static boolean overlap(int a, int b, int c, int d) {
        assert (a < b && c < d );    
        return ((c < a && a < d) || (c >= a && c < b));

    }


    @Override
    public int compareTo(Task o) {
        return this.est > o.est ? 1 : this.est < o.est ? -1 : 0;
    }

    @Override
    public String toString() {
        return "Task: (est = " + this.est + ", lct = " + this.lct + ", p = " + this.p + ")";
    }
}
